from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TypeVar, Generic, TYPE_CHECKING

from serious.descriptors import TypeDescriptor

M = TypeVar('M')  # Python model value
S = TypeVar('S')  # Serialized value

if TYPE_CHECKING:
    from .model import SeriousModel
    from .context import Loading, Dumping


class Serializer(Generic[M, S], ABC):

    @abstractmethod
    def load(self, value: S, ctx: Loading) -> M:
        raise NotImplementedError

    @abstractmethod
    def dump(self, value: M, ctx: Dumping) -> S:
        raise NotImplementedError


class FieldSerializer(Serializer[S, M], ABC):
    """
    A abstract field serializer defining a constructor invoked by serious `dump`, `load` and class `fits` methods.

    Field serializers are provided to a serious model (`JsonModel`_, `DictModel`_, `YamlModel`_) serializers
    parameter in an order in which they will be tested for fitness in.

    A clean way to add custom serializers to the defaults is to use the `field_serializers` function.

    .. _JsonModel: serious.json.model.JsonModel
    .. _DictModel: serious.dict.model.DictModel
    .. _YamlModel: serious.yaml.model.YamlModel
    """

    def __init__(self, descriptor: TypeDescriptor, root_model: 'SeriousModel'):
        self.type = descriptor
        self.root = root_model

    @classmethod
    @abstractmethod
    def fits(cls, desc: TypeDescriptor) -> bool:
        """
        A predicate returning `True` if this serializer fits to load/dump data for the provided field.

        The first fitting `FieldSerializer` from the list provided to the model will be used.

        Beware, the `field.type.cls` property can be an instance of a generic alias which will error,
        if using `issubclass` which expects a `type`.
        """
        raise NotImplementedError



"""Minor utilities used by JSON model."""

__all__ = ['camel_to_snake', 'snake_to_camel']

import re

first_cap_re = re.compile(r'(.)([A-Z][a-z]+)')
all_cap_re = re.compile(r'([a-z0-9])([A-Z])')
digit_re = re.compile(r'([a-z])([0-9])')


def camel_to_snake(camel: str) -> str:
    s1 = first_cap_re.sub(r'\1_\2', camel)
    s2 = all_cap_re.sub(r'\1_\2', s1).lower()
    return digit_re.sub(r'\1_\2', s2)


def snake_to_camel(snake: str) -> str:
    first, *others = filter(bool, snake.split('_'))
    return ''.join([first.lower(), *map(str.title, others)])



"""Map dataclass field names to different keys when serializing.

The biggest case to remap keys is switch between dataclass snake_case and JSON camelCase traditions.
We honor foreign traditions, so the JSON model does by default.

Remapping is implementing by creating a custom KeyMapper and passing it as a parameter to `SeriousModel`.
"""

__all__ = ['KeyMapper', 'NoopKeyMapper']

from abc import ABC, abstractmethod


class KeyMapper(ABC):
    """A two way mapping of field names to serialized data keys.

    Implement this abstract base class and pass it to `SeriousModel`.
    """

    @abstractmethod
    def to_model(self, key: str) -> str:
        """
        :param key: the key of a serialized object
        :return: a field name of a model
        """
        raise NotImplementedError

    @abstractmethod
    def to_serialized(self, field: str) -> str:
        """
        :param field: the field name of a model
        :return: a key of a serialized object
        """
        raise NotImplementedError


class NoopKeyMapper(KeyMapper):
    """A key mapper that does nothing."""

    def to_model(self, key: str) -> str:
        return key

    def to_serialized(self, field: str) -> str:
        return field



"""Loading/dumping context holding the current execution stack, calling serializers and validators."""
from __future__ import annotations

__all__ = ['Context', 'Loading', 'Dumping']

from abc import ABC, abstractmethod
from collections import deque
from typing import Any, TypeVar, Deque

from serious.serialization.serializer import Serializer
from serious.types import FrozenList
from serious.validation import validate

M = TypeVar('M')  # Python model value
S = TypeVar('S')  # Serialized value

SerializationStep = str


class Context(ABC):
    """An abstract base class for the serialization context.

    Serialization context is created when executing `SeriousModel` load or dump and is passed inside
    all nested serializers.

    All of serializers are called via context to include them in stack and to perform
    all the necessary validation and processing
    """
    _steps: Deque[SerializationStep]

    @property
    def stack(self) -> FrozenList[SerializationStep]:
        """The stack is included in errors, mentioning the fields, array indexes, dictionary keys, etc."""
        return FrozenList(self._steps)

    def __repr__(self):
        return f"<Context: {'.'.join(self._steps)}>"

    @abstractmethod
    def run(self, step: str, serializer: Serializer, value: Any) -> Any:
        """Execute serializer in context.

        Implementations:
        - includes the current step in the stack,
        - executes current steps serializer,
        - performs any required processing of values.

        This abstraction is needed for a straightforward custom serializer implementation.
        Extracting validation (or any other value processing) from serializers to `Context#run(...)`
        has left field serializers with a plain structure of a constructor, load, dump, and fits methods.
        """
        raise NotImplementedError


class Loading(Context):
    """Context used during **load** operations."""

    def __init__(self, *, validating: bool):
        super().__init__()
        self._steps = deque()
        self.validating = validating

    def run(self, step: str, serializer: Serializer[M, S], value: S) -> M:
        self._steps.append(step)
        result = serializer.load(value, self)
        if self.validating:
            validate(result)
        self._steps.pop()
        return result


class Dumping(Context):
    """Context used during **dump** operations."""

    def __init__(self, *, validating: bool):
        super().__init__()
        self._steps = deque()
        self.validating = validating

    def run(self, step: str, serializer: Serializer[M, S], o: M) -> S:
        self._steps.append(step)
        if self.validating:
            validate(o)
        result = serializer.dump(o, self)
        self._steps.pop()
        return result



"""A module with `DictModel` -- Serious model to transform between dataclasses and dictionaries."""
from __future__ import annotations

__all__ = ['DictModel']

from typing import TypeVar, Type, Generic, List, Collection, Dict, Iterable, Any, Union

from serious.descriptors import describe, TypeDescriptor
from serious.serialization import FieldSerializer, SeriousModel, field_serializers
from serious.utils import class_path

T = TypeVar('T')


class DictModel(Generic[T]):
    """A model converting dataclasses to dicts and back.


        :Example:

        from uuid import UUID
        from dataclasses import dataclass
        from serious import DictModel

        @dataclass
        class Robot:
            serial: UUID
            name: str

        >>> model = DictModel(Robot)
        >>> model.load({'serial': 'f3179d05-30f6-43ba-b6cb-7556af09330b', 'name': 'Caliban'})
        Robot(serial=UUID('f3179d05-30f6-43ba-b6cb-7556af09330b'), name='Caliban')
        >>> model.dump(Robot(UUID('00000000-0000-4000-0000-000002716057'), 'Bender'))
        {'serial': '00000000-0000-4000-0000-000002716057', 'name': 'Bender'}

    Check `__init__` parameters for a list of configuration options.

    `More on models in docs <https://serious.readthedocs.io/en/latest/models/>`_.
    """
    descriptor: TypeDescriptor
    serious_model: SeriousModel

    def __init__(
            self,
            cls: Type[T],
            serializers: Iterable[Type[FieldSerializer]] = field_serializers(),
            *,
            allow_any: bool = False,
            allow_missing: bool = False,
            allow_unexpected: bool = False,
            validate_on_load: bool = True,
            validate_on_dump: bool = False,
            ensure_frozen: Union[bool, Iterable[Type]] = False,
    ):
        """Initialize a dictionary model.

        :param cls: the dataclass type to load/dump.
        :param serializers: field serializer classes in an order they will be tested for fitness for each field.
        :param allow_any: `False` to raise if the model contains fields annotated with `Any`
                (this includes generics like `List[Any]`, or simply `list`).
        :param allow_missing: `False` to raise during load if data is missing the optional fields.
        :param allow_unexpected: `False` to raise during load if data contains some unknown fields.
        :param validate_on_load: to call dataclass `__validate__` method after object construction.
        :param validate_on_dump: to call object `__validate__` before dumping.
        :param ensure_frozen: `False` to skip check of model immutability; `True` will perform the check
                against built-in immutable types; a list of custom immutable types is added to built-ins.
        """
        self.cls = cls
        self.descriptor = describe(cls)
        self.serious_model = SeriousModel(
            self.descriptor,
            serializers,
            allow_any=allow_any,
            allow_missing=allow_missing,
            allow_unexpected=allow_unexpected,
            validate_on_load=validate_on_load,
            validate_on_dump=validate_on_dump,
            ensure_frozen=ensure_frozen,
        )

    def load(self, data: Dict[str, Any]) -> T:
        """Load dataclass from a dictionary."""
        return self.serious_model.load(data)

    def load_many(self, items: Iterable[Dict[str, Any]]) -> List[T]:
        """Load a list of dataclasses from a dictionary."""
        return [self.load(each) for each in items]

    def dump(self, o: T) -> Dict[str, Any]:
        """Dump a dataclasses to a dictionary."""
        return self.serious_model.dump(o)

    def dump_many(self, items: Collection[T]) -> List[Dict[str, Any]]:
        """Dump a list dataclasses to a dictionary."""
        return [self.dump(o) for o in items]

    def __repr__(self):
        path = class_path(type(self))
        if path == 'serious.dict.model.DictModel':
            path = 'serious.DictModel'
        return f'<{path}[{class_path(self.cls)}] at {hex(id(self))}>'



"""Serious is a Python dataclass model toolkit for serialization, validation, and more.

Core functionality is available for import from `serious`, and is listed here in `serious/__init__`:
models, errors, types and validation.

To provide custom field serialization use `serious.serialization`.

Test utils can be found in `serious.test_utils`.

`More on Read The Docs.`_
`Sources on GitHub.`_

.. _More on Read The Docs.: https://serious.readthedocs.io/en/latest/
.. _Sources on GitHub.: https://github.com/mdrachuk/serious
"""

from .dict import DictModel
from .errors import ModelError, ValidationError, LoadError, DumpError
from .json import JsonModel
from .types import Timestamp, Email, FrozenList, FrozenDict
from .validation import validate

__version__ = '1.2.10'
